import torch
import pytorch_lightning as L
from torch.utils.data import Dataset
from torchvision import datasets
import torch.optim as optim
import torch.nn as nn
from sklearn.manifold import TSNE
from sklearn.preprocessing import StandardScaler
from sklearn.metrics import roc_auc_score
from matplotlib import pyplot as plt
from typing import Union, Callable, Tuple, Any
import pandas as pd
import numpy as np
from collections import Counter

from .utils import add_postfix_to_id_ra_dec_col, correlate
from .connection import mark_object_in_catalog


def tsne(data: pd.DataFrame, ignored: list, verbose: bool = False, random_state: int = 42,
         perplexity: int = 400, scalar: bool = True, n_components: int = 2) -> pd.DataFrame:

    model = TSNE(n_components=n_components, learning_rate='auto', init='random', perplexity=perplexity,
                 random_state=random_state, verbose=verbose)

    data = data.drop(columns=ignored)

    for i in data.columns:
        data[i] = data[i].fillna(data.loc[data[i].notna(), i].mean())
        if data[i].isna().sum() > 0:
            del data[i]

    if scalar:
        data = StandardScaler().fit_transform(data)
    X_tr = model.fit_transform(data)
    return X_tr


def plot_hmxb(compressed: np.ndarray, train: pd.DataFrame, ax: Union[plt.Axes, None] = None, name: str = '') -> None:  # *(train.g_mag<12)

    if ax is None:
        fig, ax = plt.subplots()
    compressed_false = compressed[train.type=='None']
    ax.scatter(compressed_false[:, 0], compressed_false[:, 1], s=0.1, c='g', label='Unknown')

    color = {'CV': 'orange', 'Quasar': 'black', 'Star': 'purple','LMXB': 'b',  'HMXB': 'r', }

    for key, val in color.items():
        x = compressed[train.type.values == key]
        ax.scatter(x[:, 0], x[:, 1], s=6, c=val, label=key)

    plt.legend()


    if name:
        ax.set_title(name)

def plot_hmxb(compressed: np.ndarray, train: pd.DataFrame, ax: Union[plt.Axes, None] = None, name: str = '',) -> None:  # *(train.g_mag<12)

    if ax is None:
        fig, ax = plt.subplots()
    compressed_false = compressed[train.otype.isna()]
    ax.scatter(compressed_false[:, 0], compressed_false[:, 1], s=0.1, c='g', label='Unknown')

    if 'otype' in train.columns:
        color = {'Quasar': 'black', 'CV':'orange', 'Galaxy': 'purple', 'EB*': 'y'}
        for key, val in color.items():
            x = compressed[train.otype.values == key]
            ax.scatter(x[:, 0], x[:, 1], s=6, c=val)

    color = {'CV': 'orange', 'Quasar': 'black', 'Galaxy': 'purple','LMXB': 'b',  'HMXB': 'r', }

    for key, val in color.items():
        x = compressed[train.type.values == key]
        ax.scatter(x[:, 0], x[:, 1], s=6, c=val, label=key)



    plt.legend()


    if name:
        ax.set_title(name)


# ----------------------------------------------------------------------------------------------- #
#                                         Gat rank                                                #
# ----------------------------------------------------------------------------------------------- #


def dist_to_n_nearest_(n_coords: np.ndarray, m_coords: np.ndarray, count_nearest: int = 5, how: str = 'mean') -> np.ndarray:
    """
    ['mean', 'exp', 'harmonic', 'geometric']
    """
    nearest_distances = []
    for m_point in m_coords:
        distances = np.sqrt(np.sum((n_coords - m_point)**2, axis=1))
        nearest_indices = np.argsort(distances)[:count_nearest]
        if how == 'mean':
            mean = distances[nearest_indices].mean()
        elif how == 'exp':
            mean = np.average(distances[nearest_indices], weights=np.exp(-0.5 * np.arange(len(nearest_indices))))
        elif how == 'harmonic':
            mean = np.average(distances[nearest_indices], weights=1 / (np.arange(len(nearest_indices)) + 1))
        elif how == 'geometric':
            mean = np.average(distances[nearest_indices], weights=1 / np.log(np.arange(len(nearest_indices)) + 2))
        nearest_distances.append(mean)
    return np.array(nearest_distances)

from scipy.spatial import cKDTree

def dist_to_n_nearest(n_coords: np.ndarray, m_coords: np.ndarray, count_nearest: int = 5, how: str = 'mean') -> np.ndarray:
    """
    ['mean', 'exp', 'harmonic', 'geometric']
    """
    ntree = cKDTree(n_coords)
    nearest_distances = []

    for m_point in m_coords:
        distances, _ = ntree.query(m_point, k=count_nearest)
        
        if how == 'mean':
            mean = np.mean(distances)
        elif how == 'exp':
            mean = np.average(distances, weights=np.exp(-0.5 * np.arange(len(distances))))
        elif how == 'harmonic':
            mean = np.average(distances, weights=1 / (np.arange(len(distances)) + 1))
        elif how == 'geometric':
            mean = np.average(distances, weights=1 / np.log(np.arange(len(distances)) + 2))
        
        nearest_distances.append(mean)
    
    return np.array(nearest_distances)


# ----------------------------------------------------------------------------------------------- #
#                                  метрика отбора метода                                          #
# ----------------------------------------------------------------------------------------------- #
def _number_top_n(ranks: np.ndarray, n: int) -> int:
    return (ranks <= n).sum()


def number_top_n(sorted_data: pd.DataFrame, mark: np.ndarray, n: int) -> int:
    return _number_top_n(sorted_data[mark].index, n)


def number_top_10(sorted_data: pd.DataFrame, mark: np.ndarray) -> int:
    return _number_top_n(sorted_data[mark].index, 10)


def number_top_100(sorted_data: pd.DataFrame, mark: np.ndarray) -> int:
    return _number_top_n(sorted_data[mark].index, 100)


def roc_auc_rank(sorted_data: pd.DataFrame, mark: np.ndarray) -> float:
    return roc_auc_score(mark, -sorted_data.index)


def evaluate_compression(compress: np.ndarray, data: pd.DataFrame, samples: int = 10, sample_size: int = 20,
                         metric: str = 'mean', n_nearest: int = 10,
                         score_fn: Callable = lambda x, y: number_top_n(x, y, n=10)) -> float:
    scores = []
    for i in range(samples):
        data_ = data[['id_gaia', 'is_hmxb']].copy()
        data_['id_gaia'] = data_['id_gaia'].astype(np.int64)
        id_hmxb = data_.query('is_hmxb==1')['id_gaia'].sample(sample_size).values

        data_['metric'] = dist_to_n_nearest(compress[data_.is_hmxb.astype(bool) * (~data_.id_gaia.isin(id_hmxb))],
                                            compress, n_nearest, metric)
        data_.sort_values('metric', inplace=True)
        data_ = data_[(data_['is_hmxb'] != 1) | (data_['id_gaia'].isin(id_hmxb))].reset_index(drop=True)
        score = score_fn(data_, data_.id_gaia.isin(id_hmxb))
        scores.append(score)
    return np.array(scores).mean()


# ----------------------------------------------------------------------------------------------- #
#                                  подготовка кандидатов                                          #
# ----------------------------------------------------------------------------------------------- #

def get_metric(data: pd.DataFrame, compressed: np.ndarray, how: str = 'mean',
               n_neighbour: int = 10, metric_to_all: bool = False) -> np.ndarray:
    if metric_to_all:
        X_tr_false = compressed
    else:
        X_tr_false = compressed[(~data.is_hmxb.values.astype(bool))]
    X_tr_gaia = compressed[data.is_hmxb.values.astype(bool)]
    metric = dist_to_n_nearest(X_tr_gaia, X_tr_false, n_neighbour, how)
    return metric


def prepare_candidates(train: pd.DataFrame, metric: np.ndarray) -> pd.DataFrame:
    candidates = train[['id_gaia', 'id_xmm', 'id_ps', 'ra_gaia', 'dec_gaia', 'ra_xmm',
                        'dec_xmm', 'g_mag', 'p_quasar', 'p_galaxy', 'p_star', 'is_lmxb']].loc[~train.is_hmxb.values.astype(bool)]

    candidates['metric'] = metric
    candidates = candidates.sort_values('metric')
    candidates.reset_index(inplace=True, drop=True)
    candidates = candidates.rename(columns={'metric': 'score'})
    candidates['p_quasar'] = candidates['p_quasar'].round(2)
    candidates['p_galaxy'] = candidates['p_galaxy'].round(2)
    candidates['p_star'] = candidates['p_star'].round(2)
    candidates['score'] = candidates['score'].round(2)
    candidates['to_copy'] = candidates['ra_xmm'].astype(str) + ' ' + candidates['dec_xmm'].astype(str)
    candidates['url_xray'] = f'http://cdsportal.u-strasbg.fr/?target=' + candidates["ra_xmm"].astype(str) + '%20' +\
        candidates["dec_xmm"].astype(str)
    candidates['url_gaia'] = f'http://cdsportal.u-strasbg.fr/?target=' + candidates["ra_gaia"].astype(str) + '%20' +\
        candidates["dec_gaia"].astype(str)

    xmm_ = pd.read_pickle('DS/XMM/4XMM_raw.pickle')
    xmm_ = xmm_[['srcid', 'sc_hr1', 'sc_hr2', 'sc_hr3', 'sc_hr4', 'sc_ep_2_flux', 'sc_ep_3_flux']].rename(columns={'srcid': 'id_xmm', 'sc_hr1': 'hr1',
                                                                                                                   'sc_hr2': 'hr2', 'sc_hr3': 'hr3', 'sc_hr4': 'hr4', })
    xmm_['flux_0.5-2'] = xmm_['sc_ep_2_flux'] + xmm_['sc_ep_3_flux']
    del xmm_['sc_ep_2_flux'], xmm_['sc_ep_3_flux']
    candidates = pd.merge(candidates, xmm_, on='id_xmm', how='left')

    old = pd.read_csv('DS/find_hmxb/old_candidates.csv')
    old = old.head(50)
    candidates['was_last_top50'] = candidates.id_xmm.isin(old.id_xmm)

    candidates = candidates[['id_gaia', 'id_xmm', 'id_ps', 'ra_gaia', 'dec_gaia', 'g_mag', 'ra_xmm', 'dec_xmm',
                            'flux_0.5-2', 'hr1', 'hr2', 'hr3', 'hr4', 'p_quasar', 'p_galaxy', 'p_star',
                             'score', 'was_last_top50', 'is_lmxb', 'to_copy', 'url_xray', 'url_gaia']]

    candidates['id_gaia'] = candidates['id_gaia'].astype(str)
    candidates['id_xmm'] = candidates['id_xmm'].astype(str)
    candidates['id_ps'] = candidates['id_ps'].astype(str)
    return candidates

# ----------------------------------------------------------------------------------------------- #
#                                          Нейронки                                               #
# ----------------------------------------------------------------------------------------------- #


class DeepAutoencoder(nn.Module):
    def __init__(self, shape_0: int) -> None:
        input_dim = shape_0
        hidden_dim1 = 20
        hidden_dim2 = 10
        latent_dim = 2
        super(DeepAutoencoder, self).__init__()
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, hidden_dim1),
            nn.ReLU(),
            nn.Linear(hidden_dim1, hidden_dim2),
            nn.ReLU(),
            nn.Linear(hidden_dim2, latent_dim)
        )
        self.decoder = nn.Sequential(
            nn.Linear(latent_dim, hidden_dim2),
            nn.ReLU(),
            nn.Linear(hidden_dim2, hidden_dim1),
            nn.ReLU(),
            nn.Linear(hidden_dim1, input_dim),
            nn.Sigmoid()
        )

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        x = self.decoder(x)
        return x


class MyEncoder(nn.Module):

    def __init__(self, shapes: list, relu: bool = True, dropout: Union[float, None] = None) -> None:
        super().__init__()
        list_of_layers = []
        for i in range(len(shapes) - 2):
            list_of_layers.append(nn.Linear(shapes[i], shapes[i + 1]))
            if relu:
                list_of_layers.append(nn.ReLU())
            if dropout is not None:
                list_of_layers.append(nn.Dropout(p=dropout))
        list_of_layers.append(nn.Linear(shapes[-2], shapes[-1]))

        self.layers = nn.Sequential(*list_of_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class MyDecoder(nn.Module):

    def __init__(self, shapes: list, relu: bool = True, dropout: Union[float, None] = None) -> None:
        super().__init__()
        list_of_layers = []
        for i in range(len(shapes) - 2):
            list_of_layers.append(nn.Linear(shapes[-i - 1], shapes[-i - 2]))
            if relu:
                list_of_layers.append(nn.ReLU())
            if dropout is not None:
                list_of_layers.append(nn.Dropout(p=dropout))
        list_of_layers.append(nn.Linear(shapes[1], shapes[0]))
        self.layers = nn.Sequential(*list_of_layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.layers(x)


class MyAutoencoder(L.LightningModule):

    def __init__(self, shapes: list, relu: bool = True, dropout_en: Union[float, None] = None,
                 dropout_de: Union[float, None] = None) -> None:
        super().__init__()
        self.encoder = MyEncoder(shapes, relu, dropout_en)
        self.decoder = MyDecoder(shapes, relu, dropout_de)

        self.val_outputs: list = []
        self.train_outputs: list = []
        self.train_loss: list = []
        self.val_loss: list = []

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = self.encoder(x)
        x = self.decoder(x)
        return x

    def calc_loss(self, x: torch.Tensor, x_hat: torch.Tensor) -> torch.Tensor:
        loss = nn.functional.mse_loss(x_hat, x)
        return loss

    def training_step(self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> dict[str, torch.Tensor]:
        x, y = batch
        x = x.view(x.size(0), -1)
        y = y.view(y.size(0), -1)
        x_hat = self.forward(x)
        loss = self.calc_loss(x_hat, y)
        self.log("train_loss", loss, on_step=False, on_epoch=True, prog_bar=True, logger=False)
        self.train_outputs.append({'loss': loss})
        return {'loss': loss}

    def validation_step(self, batch: Tuple[torch.Tensor, torch.Tensor], batch_idx: int) -> dict[str, torch.Tensor]:
        with torch.no_grad():
            x, y = batch
            x = x.view(x.size(0), -1)
            x_hat = self.forward(x)
            loss = self.calc_loss(x_hat, y)
        self.log("val_loss", loss, prog_bar=True)
        self.val_outputs.append({'loss': loss})
        return {'loss': loss}

    def configure_optimizers_(self) -> Tuple[list[torch.optim.Optimizer], list[dict[str, Any]]]:
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3, weight_decay=5e-4)

        lr_scheduler = {
            'scheduler': torch.optim.lr_scheduler.ReduceLROnPlateau(optimizer, mode='min', factor=0.5, patience=10, min_lr=1e-6, verbose=True),
            'monitor': 'val_loss',
        }

        return [optimizer], [lr_scheduler]

    def configure_optimizers(self) -> Tuple[list[torch.optim.Optimizer], list[dict[str, Any]]]:
        optimizer = torch.optim.Adam(self.parameters(), lr=1e-3, weight_decay=5e-4)

        # Define a custom lambda function for the sawtooth learning rate
        def lr_lambda(epoch: int) -> float:
            # Define the sawtooth shape for the learning rate
            cycle = 10  # Number of iterations in a cycle
            return 0.5 * (1 - abs(epoch % cycle - cycle // 2) / (cycle // 2))

        lr_scheduler = {
            'scheduler': torch.optim.lr_scheduler.LambdaLR(optimizer, lr_lambda),
            'interval': 'epoch',
        }

        return [optimizer], [lr_scheduler]

    def on_validation_epoch_end(self) -> None:
        # здесь вы можете использовать self.train_outputs для доступа к выводам обучения
        avg_loss = torch.stack([x['loss'] for x in self.val_outputs]).mean()
        print(f"| Val Loss: {avg_loss:.3f}")
        self.log('val_loss', avg_loss, prog_bar=True, on_epoch=True, on_step=False)
        self.val_outputs = []
        self.val_loss.append(avg_loss)

    def on_train_epoch_end(self) -> None:
        # здесь вы можете использовать self.train_outputs для доступа к выводам обучения
        avg_loss = torch.stack([x['loss'] for x in self.train_outputs]).mean()
        print(f"| Train Loss: {avg_loss:.3f}")
        self.log('train_loss', avg_loss, prog_bar=True, on_epoch=True, on_step=False)
        self.train_outputs = []
        self.train_loss.append(avg_loss)


class MyDataset(Dataset):
    def __init__(self, dataset: np.ndarray) -> None:
        self.dataset = dataset
        self.dataset = torch.tensor(self.dataset, dtype=torch.float)

    def __len__(self) -> int:
        return len(self.dataset)

    def __getitem__(self, idx: int) -> Tuple[torch.Tensor, torch.Tensor]:
        return self.dataset[idx], self.dataset[idx]
    
# ----------------------------------------------------------------------------------------------- #
#                                          Добавить классы                                        #
# ----------------------------------------------------------------------------------------------- #


def make_star1():
    star_1 = pd.read_pickle('DS/find_hmxb/mark_in_catalog/star_1.pkl')
    star_1 = star_1[['recno', '_RA', '_DE']].rename(columns={'recno':'id', '_RA': 'ra', '_DE':'dec'})
    add_postfix_to_id_ra_dec_col(star_1, 'star1')
    star_1.attrs['name'] = 'star1'
    star_1.attrs['id'] = 'id_star1'
    star_1.attrs['ra'] = 'ra_star1'
    star_1.attrs['dec'] = 'dec_star1'

    return star_1

def make_star2():
    star_2 = pd.read_pickle('DS/find_hmxb/mark_in_catalog/star_2.pkl')
    star_2 = star_2[['recno', 'RAJ2000', 'DEJ2000']].rename(columns={'recno':'id', 'RAJ2000': 'ra', 'DEJ2000':'dec'}).loc[1:]
    add_postfix_to_id_ra_dec_col(star_2, 'star2')
    star_2.attrs['name'] = 'star2'
    star_2.attrs['id'] = 'id_star2'
    star_2.attrs['ra'] = 'ra_star2'
    star_2.attrs['dec'] = 'dec_star2'

    star_2['ra_star2'] = star_2['ra_star2'].map(lambda x: float(x.split(' ')[0])*15 + float(x.split(' ')[1])*0.25 + float(x.split(' ')[2])*(0.25 / 60))
    star_2['dec_star2'] = star_2['dec_star2'].map(lambda x: float(x.split(' ')[0]) + float(x.split(' ')[1])*(1/60) + float(x.split(' ')[2])*(1/3600))
    return star_2

def make_quas():
    quas = pd.read_pickle('DS/find_hmxb/mark_in_catalog/mil_quasar.pkl')
    quas = quas[['recno', 'RAJ2000', 'DEJ2000']].rename(columns={'recno':'id', 'RAJ2000': 'ra', 'DEJ2000':'dec'}).copy()
    add_postfix_to_id_ra_dec_col(quas, 'quas')
    quas.attrs['name'] = 'quas'
    quas.attrs['id'] = 'id_quas'
    quas.attrs['ra'] = 'ra_quas'
    quas.attrs['dec'] = 'dec_quas'

    return quas

def make_cv_cat1():
    cv_cat1 = pd.read_pickle('DS/find_hmxb/mark_in_catalog/cv_cat1.pkl')
    cv_cat1 = cv_cat1[['recno', 'RAJ2000', 'DEJ2000']].rename(columns={'recno':'id', 'RAJ2000': 'ra', 'DEJ2000':'dec'}).copy()
    add_postfix_to_id_ra_dec_col(cv_cat1, 'cv1')
    cv_cat1.attrs['name'] = 'cv1'
    cv_cat1.attrs['id'] = 'id_cv1'
    cv_cat1.attrs['ra'] = 'ra_cv1'
    cv_cat1.attrs['dec'] = 'dec_cv1'

    cv_cat1['ra_cv1'] = cv_cat1['ra_cv1'].map(lambda x: float(x.split(' ')[0])*15 + float(x.split(' ')[1])*0.25 + float(x.split(' ')[2])*(0.25 / 60))
    cv_cat1['dec_cv1'] = cv_cat1['dec_cv1'].map(lambda x: float(x.split(' ')[0]) + float(x.split(' ')[1])*(1/60) + float(x.split(' ')[2])*(1/3600))
    return cv_cat1

def make_cv_cat2():
    cv_cat2 = pd.read_pickle('DS/find_hmxb/mark_in_catalog/cv_cat2.pkl')
    cv_cat2 = cv_cat2[['recno', 'RAJ2000', 'DEJ2000']].rename(columns={'recno':'id', 'RAJ2000': 'ra', 'DEJ2000':'dec'}).copy()
    add_postfix_to_id_ra_dec_col(cv_cat2, 'cv2')
    cv_cat2.attrs['name'] = 'cv2'
    cv_cat2.attrs['id'] = 'id_cv2'
    cv_cat2.attrs['ra'] = 'ra_cv2'
    cv_cat2.attrs['dec'] = 'dec_cv2'

    cv_cat2['ra_cv2'] = cv_cat2['ra_cv2'].map(lambda x: float(x.split(' ')[0])*15 + float(x.split(' ')[1])*0.25 + float(x.split(' ')[2])*(0.25 / 60))
    cv_cat2['dec_cv2'] = cv_cat2['dec_cv2'].map(lambda x: float(x.split(' ')[0]) + float(x.split(' ')[1])*(1/60) + float(x.split(' ')[2])*(1/3600))
    return cv_cat2

def make_lis():
    desi_lis = pd.read_pickle('DS/LS/LS_class.pickle')
    add_postfix_to_id_ra_dec_col(desi_lis, 'lis')
    desi_lis.attrs['name'] = 'lis'
    desi_lis.attrs['id'] = 'id_lis'
    desi_lis.attrs['ra'] = 'ra_lis'
    desi_lis.attrs['dec'] = 'dec_lis'

    return desi_lis

def add_hmxb_and_lmxb(df_main):
    hmxb = pd.read_pickle('DS/HMXB/HMXB_2023.pickle')
    lmxb = pd.read_pickle('DS/HMXB/LMXB_2023.pickle')
    df_main['is_hmxb'] = df_main['id_gaia'].isin(hmxb['Gaia_DR3_ID']).astype(int)
    df_main['is_lmxb'] = df_main['id_gaia'].isin(lmxb['Gaia_DR3_ID']).astype(int)

    return df_main

def add_classes_real(df_main):
    quas = make_quas()
    star1 = make_star1()
    star2 = make_star2()
    cv1 = make_cv_cat1()
    cv2 = make_cv_cat2()
    lis = make_lis()

    df_main = mark_object_in_catalog(df_main, quas, 0.2)
    df_main = mark_object_in_catalog(df_main, cv1, 3)
    df_main = mark_object_in_catalog(df_main, cv2, 3)
    df_main = mark_object_in_catalog(df_main, star1, 5)
    df_main = mark_object_in_catalog(df_main, star2, 1)

    cor = correlate(df_main, lis, 0.5, add_distance=True)

    df_main = add_hmxb_and_lmxb(df_main)

    df_main['type'] = 'None'
    df_main.loc[df_main.is_in_quas==True, 'type'] = 'Quasar'
    df_main.loc[df_main.is_in_cv1==True, 'type'] = 'CV'
    df_main.loc[df_main.is_in_cv2==True, 'type'] = 'CV'
    df_main.loc[df_main.is_in_star1==True, 'type'] = 'Star'
    df_main.loc[df_main.is_in_star2==True, 'type'] = 'Star'
    df_main.loc[df_main.id_gaia.isin(cor.query('spectype==0')['id_gaia']), 'type'] = 'Quasar'
    df_main.loc[df_main.id_gaia.isin(cor.query('spectype==1')['id_gaia']), 'type'] = 'Galaxy'
    df_main.loc[df_main.id_gaia.isin(cor.query('spectype==2')['id_gaia']), 'type'] = 'Star'
    df_main.loc[df_main.is_hmxb==True, 'type'] = 'HMXB'
    df_main.loc[df_main.is_lmxb==True, 'type'] = 'LMXB'

    del df_main['is_in_quas'], df_main['is_in_cv1'], df_main['is_in_cv2'], df_main['is_in_star1'],\
          df_main['is_in_star2'], df_main['is_hmxb'], df_main['is_lmxb']

    return df_main


# ----------------------------------------------------------------------------------------------- #
#                                          Clastering                                             #
# ----------------------------------------------------------------------------------------------- #


def prepare_classes(compressed, data, model, ax=None, name=''):
    if ax is None:
        fig, ax = plt.subplots()
    X = compressed[data.type=='HMXB']
    classes = model.fit_predict(X)
    while classes[0]!=0 or classes[6]!=2 or classes[8]!=3:
        classes = model.fit_predict(X)
    color = ['red', 'blue', 'green', 'purple', 'yellow', 'pink']
    ax.scatter(compressed[:,0], compressed[:,1], s=0.001)
    ax.scatter(X[:,0], X[:,1], color=[color[i] for i in classes])
    classes = np.array([i if i!=-1 else np.max(classes)+1 for i in classes])
    ax.set_title(name)
    return classes

def plot_class_distribution(classes, data, hmxb, ax=None, n_classes=5, name=''):
    if ax is None:
        fig, ax = plt.subplots()

    class_to_id_gaia = lambda x: data.query('type=="HMXB"').loc[classes==x].id_gaia

    category_counts = Counter()
    for i in range(n_classes):
        id_gaia = class_to_id_gaia(i)
        hmxb_classes = ','.join(map(str, list(hmxb.query("id_gaia in @id_gaia").Xray_Type.values))).split(',')
        category_counts.update(hmxb_classes)
    
    sorted_categories = [cat for cat, _ in category_counts.most_common()]

    # Построение гистограммы с отсортированными категориями
    color = ['red', 'blue', 'green', 'purple', 'yellow', 'pink']  # замените на ваши цвета
    bottom = np.zeros(len(sorted_categories))

    for i in range(n_classes):  # замените на ваш реальный диапазон классов
        id_gaia = class_to_id_gaia(i)
        hmxb_classes = ','.join(map(str, list(hmxb.query("id_gaia in @id_gaia").Xray_Type.values))).split(',')
        counts = dict(Counter(hmxb_classes))
        values = [counts.get(cat, 0) for cat in sorted_categories]

        ax.bar(sorted_categories, values, bottom=bottom, color=color[i])
        bottom += np.array(values)
    ax.set_title(name)

    
def plot_class_distribution_separately(classes, data, hmxb, n_classes=5, names=['','','','',''], axes=None):
    if axes is None:
        fig, axes = plt.subplots(1, 5, figsize=(20, 4))

    class_to_id_gaia = lambda x: data.query('type=="HMXB"').loc[classes==x].id_gaia

    category_counts_all = Counter()
    for i in range(n_classes):
        id_gaia = class_to_id_gaia(i)
        hmxb_classes = ','.join(map(str, list(hmxb.query("id_gaia in @id_gaia").Xray_Type.values))).split(',')
        category_counts_all.update(hmxb_classes)

    sorted_categories = list(category_counts_all.keys())

    for i, ax in enumerate(axes):
        category_counts = Counter()
        id_gaia = class_to_id_gaia(i)
        hmxb_classes = ','.join(map(str, list(hmxb.query("id_gaia in @id_gaia").Xray_Type.values))).split(',')
        category_counts.update(hmxb_classes)

        color = ['red', 'blue', 'green', 'purple', 'yellow', 'pink']  # замените на ваши цвета
        bottom = np.zeros(len(sorted_categories))

        values = [category_counts.get(cat, 0) for cat in sorted_categories]

        ax.bar(sorted_categories, values, bottom=bottom, color=color[i])
        ax.set_title(f'Class {i}')
        ax.set_xlabel('HMXB Type')
        ax.set_ylabel('Count')
        ax.set_xticklabels(list(sorted_categories), rotation=90)
        ax.set_title(names[i])


def calc_centers(classes, compressed, data):
    coords = compressed[data['type']=='HMXB']
    centers = {}
    for class_ in set(classes):
        centers[class_] = np.mean(coords[classes==class_], axis=0)
    return centers

def plot_centers(centers, ax=None, name=''):
    if ax is None:
        fig, ax = plt.subplots()
    for i in centers:
        ax.scatter(centers[i][0], centers[i][1], color='black', s=30)
    ax.set_title(name)  


#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import print_function, division
import nwaylib.checkupdates

__doc__ = """Multiway association between astrometric catalogue. Use --help for usage.

Example: nway.py --radius 10 --prior-completeness 0.95 --mag GOODS:mag_H auto --mag IRAC:mag_irac1 auto cdfs4Ms_srclist_v3.fits :Pos_error CANDELS_irac1.fits 0.5 gs_short.fits 0.1 --out=out.fits
"""

import sys
import numpy
from numpy import log10, pi
import astropy.io.fits as pyfits
import argparse
import tqdm
import nwaylib.progress as progress
import nwaylib.logger as logger
import nwaylib.fastskymatch as match
import nwaylib.bayesdistance as bayesdist
import nwaylib.magnitudeweights as magnitudeweights


def make_errors_table_matrix(table_names, pos_errors):
    symmetric = True
    rotated = False
    errors = []
    for ti, (table_name, pos_error) in enumerate(zip(table_names, pos_errors)):
        colnames = tables[ti].dtype.names
        if pos_error[0] != ':':
            print('    Position error for "%s": using fixed value %f' % (table_name, float(pos_error)))
            pos_error = float(pos_error)
            if pos_error > match_radius * 60 * 60:
                print('WARNING: Given separation error for "%s" is larger than the match radius! Increase --radius to >> %s' % (table_name, pos_error))
            pos_error = float(pos_error) * numpy.ones(len(table))
            errors.append((pos_error, pos_error, numpy.zeros_like(pos_error)))
            continue

        keys = pos_error[1:].split(':')
        if len(keys) > 3:
            assert False, 'Invalid column specifier: %s' % pos_error

        meanings = ['ra_error', 'dec_error', 'ell_angle']
        for k, meaning in zip(keys, meanings):
            k2 = "%s_%s" % (table_name, k)
            # get column
            assert k2 in table.dtype.names, 'ERROR: Position error column "%s" not in table "%s". Have these columns: %s' % (k2, table_name, ', '.join(colnames))
            print('    Position error for "%s": found column %s (for %s): Values are [%f..%f]' % (
                table_name, k, meaning, tables[ti][k].min(), tables[ti][k].max()))

        if len(keys) == 3:
            keys = keys[0], keys[1], keys[2]
            rotated = True

            intable_errors_ra = tables[ti][keys[0]]
            intable_errors_dec = tables[ti][keys[1]]
            intable_errors_rho = tables[ti][keys[2]] / 180 * pi

            table_errors_ra = table["%s_%s" % (table_name, keys[0])]
            table_errors_dec = table["%s_%s" % (table_name, keys[1])]
            table_errors_rho = table["%s_%s" % (table_name, keys[2])] / 180 * pi

            intable_errors_ra, intable_errors_dec, intable_errors_rho = bayesdist.convert_from_ellipse(intable_errors_ra, intable_errors_dec, intable_errors_rho)
            table_errors_ra, table_errors_dec, table_errors_rho = bayesdist.convert_from_ellipse(table_errors_ra, table_errors_dec, table_errors_rho)

        elif len(keys) == 2:
            keys = keys[0], keys[1]
            symmetric = False

            intable_errors_ra = tables[ti][keys[0]]
            intable_errors_dec = tables[ti][keys[1]]
            intable_errors_rho = numpy.zeros_like(intable_errors_ra)

            table_errors_ra = table["%s_%s" % (table_name, keys[0])]
            table_errors_dec = table["%s_%s" % (table_name, keys[1])]
            table_errors_rho = numpy.zeros_like(table_errors_ra)

        elif len(keys) == 1:
            keys = keys[0], keys[0]

            intable_errors_ra = tables[ti][keys[0]]
            intable_errors_dec = intable_errors_ra
            intable_errors_rho = numpy.zeros_like(intable_errors_ra)

            table_errors_ra = table["%s_%s" % (table_name, keys[0])]
            table_errors_dec = table_errors_ra
            table_errors_rho = numpy.zeros_like(table_errors_ra)

        if intable_errors_ra.min() <= 0 or intable_errors_dec.min() <= 0:
            print('WARNING: Some separation errors in "%s" are 0! This will give invalid results (%d rows).' % (
                keys[0], numpy.logical_and(intable_errors_ra <= 0, intable_errors_dec <= 0).sum()))
        if intable_errors_ra.max() > match_radius * 60 * 60 or intable_errors_dec.max() > match_radius * 60 * 60:
            print('WARNING: Some separation errors in "%s" are larger than the match radius! Increase --radius to >> %s' % (keys[0], max(intable_errors_ra.max(), intable_errors_dec.max())))

        errors.append((table_errors_ra, table_errors_dec, table_errors_rho))
    return errors, (not rotated) and symmetric

# set up program arguments


class HelpfulParser(argparse.ArgumentParser):
    def error(self, message):
        sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)


parser = HelpfulParser(description=__doc__,
                       epilog="""Johannes Buchner (C) 2013-2017 <johannes.buchner.acad@gmx.com>""",
                       formatter_class=argparse.ArgumentDefaultsHelpFormatter)

parser.add_argument('--radius', type=float, required=True,
                    help='exclusive search radius in arcsec for initial matching')

parser.add_argument('--mag-radius', default=None, type=float,
                    help='search radius for building the magnitude histogram of target sources. If not set, the Bayesian posterior is used.')

parser.add_argument('--mag-auto-minprob', default=0.9, type=float,
                    help='minimum posterior probability (default: 0.9) for the magnitude histogram of secure target sources. Used in the Bayesian procedure.''')

parser.add_argument('--mag-exclude-radius', default=None, type=float,
                    help='exclusion radius for building the magnitude histogram of field sources. If not set, --mag-radius is used.')

parser.add_argument('--prior-completeness', metavar='COMPLETENESS', default="1", type=str,
                    help='expected matching completeness of sources (prior)')

parser.add_argument('--ignore-unrelated-associations', dest='consider_unrelated_associations', action='store_false',
                    help='Ignore in the calculation source pairings unrelated to the primary source (not recommended)')

parser.set_defaults(consider_unrelated_associations=True)

parser.add_argument('--mag', metavar='MAGCOLUMN+MAGFILE', type=str, nargs=2, action='append', default=[],
                    help="""name of <table>:<column> for magnitude biasing, and filename for magnitude histogram
    (use auto for auto-computation within mag-radius).
    Example: --mag GOODS:mag_H auto --mag IRAC:mag_irac1 irac_histogram.txt""")

parser.add_argument('--acceptable-prob', metavar='PROB', type=float, default=0.5,
                    help='ratio limit up to which secondary solutions are flagged')

parser.add_argument('--min-prob', type=float, default=0,
                    help='lowest probability allowed in final catalogue. If 0, no trimming is performed (default).')

parser.add_argument('--out', metavar='OUTFILE', help='output file name', required=True)

parser.add_argument('catalogues', type=str, nargs='+',
                    help="""input catalogue fits files and position errors.\n
                    Example: cdfs4Ms_srclist_v3.fits :Pos_error CANDELS_irac1.fits 0.5 gs_short.fits 0.1
    """)

parser.add_argument('--prefilter-pair', metavar='CATNAME1 CATNAME2 radius', type=str, nargs=3, action='append', default=[],
                    help="""name of two <table>s where combinations more distant than radius (in arcsec)
    should not considered. This reduces the memory needs when several large catalogs
    with high accuracy are matched against some with low accuracy (high --radius).

    Example: --prefilter-pair GOODS IRAC 0.1""")


# parsing arguments
args = parser.parse_args()

print('NWAY arguments:')

diff_secondary = args.acceptable_prob
outfile = args.out

filenames = args.catalogues[::2]
print('    catalogues: ', ', '.join(filenames))
pos_errors = args.catalogues[1::2]
print('    position errors/columns: ', ', '.join(pos_errors))

fits_tables = []
table_names = []
tables = []
source_densities = []
source_densities_plus = []
fits_formats = []
for fitsname in filenames:
    fits_table = pyfits.open(fitsname)[1]
    fits_tables.append(fits_table)
    table_name = fits_table.name
    table_names.append(table_name)
    table = fits_table.data
    fits_formats.append([c.format for c in fits_table.columns])
    tables.append(table)

    n = len(table)
    assert 'SKYAREA' in fits_table.header, 'file "%s", table "%s" does not have a field "SKYAREA", which should contain the area of the catalogue in square degrees' % (fitsname, table_name)
    area = fits_table.header['SKYAREA'] * 1.0  # in square degrees
    area_total = (4 * pi * (180 / pi)**2)
    density = n / area * area_total
    print('      from catalogue "%s" (%d), density gives %.2e on entire sky' % (table_name, n, density))
    # this takes into account that the source may be absent
    density_plus = (n + 1) / area * area_total
    source_densities.append(density)
    source_densities_plus.append(density_plus)

# source can not be absent in primary catalogue
source_densities_plus[0] = source_densities[0]
source_densities_plus = numpy.array(source_densities_plus)

if ':' in args.prior_completeness:
    prior_completeness = numpy.array([1.0] + [float(pc) for pc in args.prior_completeness.split(':')])
    if len(prior_completeness) != len(filenames):
        raise Exception('Prior completeness needs one value per catalog, like "%s". Received "%s".' % (':'.join(["0.9"] * (len(filenames) - 1)), args.prior_completeness))
else:
    prior_completeness = numpy.array([1.0] + [float(args.prior_completeness)**(1. / (len(filenames) - 1)) for i in range(1, len(filenames))])


min_prob = args.min_prob

match_radius = args.radius / 60. / 60  # in degrees

pairwise_errs = [(table_names.index(tablea), table_names.index(tableb), float(err))
                 for tablea, tableb, err in args.prefilter_pair]

if len(pairwise_errs) > 0:
    print('    pair-wise pre-filtering on')

# if args.mag_radius is not None:
# mag_radius = match_radius # in arc sec
# else:
mag_include_radius = args.mag_radius  # in arc sec
mag_exclude_radius = args.mag_exclude_radius  # in arc sec
if mag_exclude_radius is None:
    mag_exclude_radius = mag_include_radius

magauto_post_single_minvalue = args.mag_auto_minprob
assert 0 < magauto_post_single_minvalue <= 1, 'probability should be between 0 and 1'

magnitude_columns = args.mag
print('    magnitude columns: ', ', '.join([c for c, _ in magnitude_columns]))

for mag, magfile in magnitude_columns:
    table_name, col_name = mag.split(':', 1)
    assert table_name in table_names, 'table name specified for magnitude ("%s") unknown. Known tables: %s' % (table_name, ', '.join(table_names))
    ti = table_names.index(table_name)
    col_names = tables[ti].dtype.names
    assert col_name in col_names, 'column name specified for magnitude ("%s") unknown. Known columns in table "%s": %s' % (mag, table_name, ', '.join(col_names))

simple_errors = True
for ti, (table_name, pos_error) in enumerate(zip(table_names, pos_errors)):
    colnames = tables[ti].dtype.names
    if pos_error[0] != ':':
        continue

    keys = pos_error[1:].split(':')
    if len(keys) > 3:
        assert False, 'Invalid column specifier: %s' % pos_error
    if len(keys) > 1:
        simple_errors = False

    meanings = [
        [],
        ['symmerror'],
        ['ra_error', 'dec_error'],
        ['majaxis', 'minaxis', 'ell_angle'],
    ][len(keys)]
    for k, meaning in zip(keys, meanings):
        # get column
        assert k in colnames, 'ERROR: Position error column "%s" not in table "%s". Have these columns: %s' % (k, table_name, ', '.join(colnames))


# first match input catalogues, compute possible combinations in match_radius
results, columns, match_header = match.match_multiple(tables, table_names, match_radius, fits_formats, circular=simple_errors,
                                                      logger=logger.NormalLogger(), pairwise_errs=pairwise_errs)
table = match.fits_from_columns(pyfits.ColDefs(columns)).data

assert len(table) > 0, 'No matches.'

# first pass: find secure matches and secure non-matches
print('Computing distance-based probabilities ...')

print('  finding position error columns ...')
# get the separation and error columns for the bayesian weighting

errors, simple_errors = make_errors_table_matrix(table_names, pos_errors)
if simple_errors:
    errors = [e_ra for e_ra, e_dec, e_rho in errors]

print('  finding position columns ...')
# table is in arcsec, and therefore separations is in arcsec


def make_separation_table_matrix(kstr, table, table_names):
    separations = []
    for ti, a in enumerate(table_names):
        row = []
        for tj, b in enumerate(table_names):
            if ti < tj:
                k = kstr % (b, a)
                assert k in table.dtype.names, 'ERROR: Separation column for "%s" not in merged table. Have columns: %s' % (k, ', '.join(table.dtype.names))
                row.append(table[k])
            else:
                row.append(numpy.ones(len(table)) * numpy.nan)
        separations.append(row)
    return separations


separations = make_separation_table_matrix('Separation_%s_%s', table, table_names)
if not simple_errors:
    separations_ra = make_separation_table_matrix('Separation_%s_%s_ra', table, table_names)
    separations_dec = make_separation_table_matrix('Separation_%s_%s_dec', table, table_names)

print('  building primary_id index ...')
primary_id_key = match.get_tablekeys(tables[0], 'ID', tablename=table_names[0])
assert len(numpy.unique(tables[0][primary_id_key])) == len(tables[0][primary_id_key]), "ERROR: ID column '%s' in primary catalog contains duplicates." % primary_id_key
primary_id_key = '%s_%s' % (table_names[0], primary_id_key)

primary_ids = []
primary_id_start = []
last_primary_id = None
primary_id_column = table[primary_id_key]
for i, pid in enumerate(primary_id_column):
    if pid != last_primary_id:
        last_primary_id = pid
        primary_ids.append(pid)
        primary_id_start.append(i)

primary_id_end = primary_id_start[1:] + [len(primary_id_column)]

# compute n-way position evidence
print('  computing probabilities ...')

log_bf = numpy.zeros(len(table)) * numpy.nan
prior = numpy.zeros(len(table)) * numpy.nan
# handle all cases (also those with missing counterparts in some catalogues)
for case in range(2**(len(table_names) - 1)):
    table_mask = numpy.array([True] + [(case // 2**(ti)) % 2 == 0 for ti in range(len(tables) - 1)])
    ncat = table_mask.sum()
    # select those cases
    mask = True
    for i in range(1, len(tables)):
        if table_mask[i]:  # require not nan
            mask = numpy.logical_and(mask, ~numpy.isnan(separations[0][i]))
        else:
            mask = numpy.logical_and(mask, numpy.isnan(separations[0][i]))
    # select errors
    if simple_errors:
        errors_selected = [e[mask] for e, m in zip(errors, table_mask) if m]
        separations_selected = [[cell[mask] for cell, m in zip(row, table_mask) if m]
                                for row, m in zip(separations, table_mask) if m]
        log_bf[mask] = bayesdist.log_bf(separations_selected, errors_selected)
    else:
        errors_selected = [(era[mask], edec[mask], ephi[mask])
                           for (era, edec, ephi), m in zip(errors, table_mask) if m]
        separations_selected_ra = [[cell[mask] for cell, m in zip(row, table_mask) if m]
                                   for row, m in zip(separations_ra, table_mask) if m]
        separations_selected_dec = [[cell[mask] for cell, m in zip(row, table_mask) if m]
                                    for row, m in zip(separations_dec, table_mask) if m]
        log_bf[mask] = bayesdist.log_bf_elliptical(
            separations_selected_ra, separations_selected_dec, errors_selected)

    prior[mask] = source_densities[0] * numpy.product(prior_completeness[table_mask]) / numpy.product(source_densities_plus[table_mask])
    assert numpy.isfinite(prior[mask]).all(), (source_densities, prior_completeness[table_mask], numpy.product(source_densities_plus[table_mask]))

assert numpy.isfinite(prior).all(), (prior, log_bf)
assert numpy.isfinite(log_bf).all(), (prior, log_bf)
columns.append(pyfits.Column(name='dist_bayesfactor', format='E', array=log_bf))

ncat = table['ncat']
ncats = len(tables)

if args.consider_unrelated_associations:
    candidates = numpy.where(ncat <= ncats - 2)[0]
    if len(candidates) > 0:
        print('    correcting for unrelated associations ...')
        # correct for unrelated associations
        # identify those in need of correction
        # two unconsidered catalogues are needed for an unrelated association
        for i in tqdm.tqdm(candidates):
            # list which ones we are missing
            missing_cats = [k for k, sep in enumerate(separations[0]) if numpy.isnan(sep[i])]
            pid = table[primary_id_key][i]
            pid_index = primary_ids.index(pid)
            best_logpost = 0
            # go through more complex associations
            for j in range(primary_id_start[pid_index], primary_id_end[pid_index]):
                if not (ncat[j] > 2):
                    continue
                # check if this association has sufficient overlap with the one we are looking for
                # it must contain at least two of the catalogues we are missing
                augmented_cats = []
                for k in missing_cats:
                    if not numpy.isnan(separations[0][k][j]):
                        augmented_cats.append(k)
                n_augmented_cats = len(augmented_cats)
                if n_augmented_cats >= 2:
                    # ok, this is helpful.
                    # identify the separations and errors
                    # identify the prior
                    prior_j = source_densities[augmented_cats[0]] / numpy.product(source_densities_plus[augmented_cats])
                    # compute a log_bf
                    if simple_errors:
                        errors_selected = [[errors[k][j]] for k in augmented_cats]
                        separations_selected = [[[separations[k][k2][j]]
                                                 for k2 in augmented_cats] for k in augmented_cats]
                        log_bf_j = bayesdist.log_bf(numpy.array(separations_selected),
                                                    numpy.array(errors_selected))
                    else:
                        separations_selected_ra = [[[separations_ra[k][k2][j]]
                                                    for k2 in augmented_cats] for k in augmented_cats]
                        separations_selected_dec = [[[separations_dec[k][k2][j]]
                                                     for k2 in augmented_cats] for k in augmented_cats]
                        errors_selected = [([errors[k][0][j]], [errors[k][1][j]], [errors[k][2][j]])
                                           for k in augmented_cats]
                        log_bf_j = bayesdist.log_bf_elliptical(numpy.array(separations_selected_ra),
                                                               numpy.array(separations_selected_dec),
                                                               numpy.array(errors_selected))
                    logpost_j = bayesdist.unnormalised_log_posterior(prior_j, log_bf_j, n_augmented_cats)
                    if logpost_j > best_logpost:
                        # print('post:', logpost_j, log_bf_j, prior_j)
                        best_logpost = logpost_j

            # ok, we have our correction factor, best_logpost
            # lets multiply it onto log_bf
            if best_logpost > 0:
                log_bf[i] += best_logpost
        columns.append(pyfits.Column(name='dist_bayesfactor_corrected', format='E', array=log_bf))
    else:
        print('      correcting for unrelated associations ... not necessary')

# add the additional columns
post = bayesdist.posterior(prior, log_bf)
columns.append(pyfits.Column(name='dist_post', format='E', array=post))

# find magnitude biasing functions
if magnitude_columns:
    print()
    print('Incorporating magnitude biases ...')
biases = {}
for mag, magfile in magnitude_columns:
    print('    magnitude bias "%s" ...' % mag)
    table_name, col_name = mag.split(':', 1)
    assert table_name in table_names, 'table name specified for magnitude ("%s") unknown. Known tables: %s' % (table_name, ', '.join(table_names))
    ti = table_names.index(table_name)
    col_names = tables[ti].dtype.names
    assert col_name in col_names, 'column name specified for magnitude ("%s") unknown. Known columns in table "%s": %s' % (mag, table_name, ', '.join(col_names))
    ci = col_names.index(col_name)

    res = results[table_name]
    res_defined = results[table_name] != -1

    # get magnitudes of all
    mag_all = tables[ti][col_name]
    # mark -99 as undefined
    mag_all[mag_all == -99] = numpy.nan

    # get magnitudes of selected
    mask_all = ~numpy.logical_or(numpy.isnan(mag_all), numpy.isinf(mag_all))

    col = "%s_%s" % (table_name, col_name)

    if magfile == 'auto':
        if mag_include_radius is not None:
            if mag_include_radius >= match_radius * 60 * 60:
                print('WARNING: magnitude radius is very large (>= matching radius). Consider using a smaller value.')
            selection = table['Separation_max'] < mag_include_radius
            selection_possible = table['Separation_max'] < mag_exclude_radius
            selection_weights = numpy.ones(len(selection))
        else:
            selection = post > magauto_post_single_minvalue
            selection_weights = post
            selection_possible = post > 0.01

        # ignore cases where counterpart is missing
        assert res_defined.shape == selection.shape, (res_defined.shape, selection.shape)
        selection = numpy.logical_and(selection, res_defined)
        selection_weights = selection_weights[selection]
        selection_possible = numpy.logical_and(selection_possible, res_defined)

        # print '   selection', selection.sum(), selection_possible.sum(), (-selection_possible).sum()

        # rows = results[table_name][selection].tolist()
        rows, unique_indices = numpy.unique(results[table_name][selection], return_index=True)
        rows_weights = selection_weights[unique_indices]

        assert len(rows) > 1, 'No magnitude values within radius for "%s".' % mag
        mag_sel = mag_all[rows]
        mag_sel_weights = rows_weights

        # remove vaguely possible options from alternative histogram
        rows_possible = numpy.unique(results[table_name][selection_possible])
        mask_others = mask_all.copy()
        mask_others[rows_possible] = False

        # all options in the total (field+target sources) histogram
        mask_sel = ~numpy.logical_or(numpy.isnan(mag_sel), numpy.isinf(mag_sel))

        # print '      non-nans: ', mask_sel.sum(), mask_others.sum()

        print('    magnitude histogram of column "%s": %d secure matches, %d insecure matches and %d secure non-matches of %d total entries (%d valid)' %
              (col, mask_sel.sum(), len(rows_possible), mask_others.sum(), len(mag_all), mask_all.sum()))

        # make function fitting to ratio shape
        bins, hist_sel, hist_all = magnitudeweights.adaptive_histograms(mag_all[mask_others], mag_sel[mask_sel], weights=mag_sel_weights[mask_sel])
        print('    magnitude histogram stored to "%s".' % (mag.replace(':', '_') + '_fit.txt'))
        with open(mag.replace(':', '_') + '_fit.txt', 'wb') as f:
            f.write(b'# lo hi selected others\n')
            numpy.savetxt(f,
                          numpy.transpose([bins[:-1], bins[1:], hist_sel, hist_all]),
                          fmt=["%10.5f"] * 4)
        if mask_sel.sum() < 100:
            print('ERROR: too few secure matches to make a good histogram. If you are sure you want to use this poorly sampled histogram' +
                  'replace "auto" with the filename. You can also decrease the mag-auto-minprob parameter.''')
            sys.exit(1)
    else:
        print('    magnitude histogramming: using histogram from "%s" for column "%s"' % (magfile, col))
        bins_lo, bins_hi, hist_sel, hist_all = numpy.loadtxt(magfile).transpose()
        bins = numpy.array(list(bins_lo) + [bins_hi[-1]])
    func = magnitudeweights.fitfunc_histogram(bins, hist_sel, hist_all)
    magnitudeweights.plot_fit(bins, hist_sel, hist_all, func, mag)
    weights = log10(func(table[col]))
    # undefined magnitudes do not contribute
    weights[numpy.isnan(weights)] = 0
    biases[col] = weights


# add the bias columns
for col, weights in biases.items():
    columns.append(pyfits.Column(name='bias_%s' % col, format='E', array=10**weights))

print()
print('Computing final probabilities ...')

# add the posterior column
total = log_bf + sum(biases.values())
post = bayesdist.posterior(prior, total)
columns.append(pyfits.Column(name='p_single', format='E', array=post))

# compute weights for group posteriors
# 4pi comes from Eq.
log_post_weight = bayesdist.unnormalised_log_posterior(prior, total, ncat)

# flagging of solutions. Go through groups by primary id (IDs in first catalogue)
index = numpy.zeros_like(post)
prob_has_match = numpy.zeros_like(post)
prob_this_match = numpy.zeros_like(post)

match_header['COL_PRIM'] = primary_id_key
match_header['COLS_ERR'] = ' '.join(['%s_%s' % (ti, poscol) for ti, poscol in zip(table_names, pos_errors)])
print('    grouping by column "%s" and flagging ...' % (primary_id_key))

pid_index = primary_ids.index(pid)
best_log_bf = 0
# go through more complex associations
for primary_id, ilo, ihi in tqdm.tqdm(list(zip(primary_ids, primary_id_start, primary_id_end))):
    # group
    mask = slice(ilo, ihi)

    # compute no-match probability
    values = log_post_weight[mask]
    offset = values.max()
    bfsum = log10((10**(values - offset)).sum()) + offset
    if len(values) > 1:
        offset = values[1:].max()
        bfsum1 = log10((10**(values[1:] - offset)).sum()) + offset
    else:
        bfsum1 = 0

    # for p_any, find the one without counterparts
    p_none = float(values[0])
    p_any = 1 - 10**(p_none - bfsum)
    # this avoids overflows in the no-counterpart solution,
    # which we want to set to 0
    values[0] = bfsum1
    p_i = 10**(values - bfsum1)
    p_i[0] = 0

    prob_has_match[mask] = p_any
    prob_this_match[mask] = p_i

    best_val = p_i.max()

    # flag best & second best
    # ignore very poor solutions
    index[mask] = numpy.where(best_val == p_i, 1,
                              numpy.where(p_i > diff_secondary * best_val, 2, 0))


columns.append(pyfits.Column(name='p_any', format='E', array=prob_has_match))
columns.append(pyfits.Column(name='p_i', format='E', array=prob_this_match))

# index[ncat == 1] == -1
# add the flagging column
columns.append(pyfits.Column(name='match_flag', format='I', array=index))

# cut away poor posteriors if requested
if min_prob > 0:
    mask = ~(prob_this_match < min_prob)
    print('    cutting away %d (below p_i minimum)' % (len(mask) - mask.sum()))

    for c in columns:
        c.array = c.array[mask]

if not filenames[0].endswith('shifted.fits'):
    print()
    print()
    print('  You can calibrate a p_any cut-off with the following steps:')
    print('   1) Create a offset catalogue to simulate random sky positions:')
    shiftfile = filenames[0].replace('.fits', '').replace('.FITS', '') + '-fake.fits'
    shiftoutfile = outfile + '-fake.fits'
    print('      nway-create-fake-catalogue.py --radius %d %s %s' % (args.radius * 2, filenames[0], shiftfile))
    print('   2) Match the offset catalogue in the same way as this run:')
    newargv = []
    i = 0
    while i < len(sys.argv):
        v = sys.argv[i]
        if v == filenames[0]:
            newargv.append(shiftfile)
        elif v == '--mag':
            newargv.append(v)
            v = sys.argv[i + 1]
            newargv.append(v)
            if sys.argv[i + 2] == 'auto':
                newargv.append(v.replace(':', '_') + '_fit.txt')
            else:
                newargv.append(sys.argv[i + 2])
            i = i + 2
        elif v == '--out':
            newargv.append(v)
            i = i + 1
            newargv.append(shiftoutfile)
        elif v.startswith('--out='):
            newargv.append('--out=' + shiftoutfile)
        else:
            newargv.append(v)
        i = i + 1
    print('      ' + ' '.join(newargv))
    print('   3) determining the p_any cutoff that corresponds to a false-detection rate')
    print('      nway-calibrate-cutoff.py %s %s' % (outfile, shiftoutfile))
    print()

# write out fits file
print()
print('creating output FITS file ...')
tbhdu = match.fits_from_columns(pyfits.ColDefs(columns))

hdulist = match.wraptable2fits(tbhdu, 'NWAYMATCH')
hdulist[0].header['METHOD'] = 'NWAY multi-way matching'
hdulist[0].header['INPUT'] = ', '.join(filenames)
hdulist[0].header['TABLES'] = ', '.join(table_names)
hdulist[0].header['BIASING'] = ', '.join(biases.keys())
hdulist[0].header['NWAYCMD'] = ' '.join(sys.argv)
for k, v in args.__dict__.items():
    hdulist[0].header.add_comment("argument %s: %s" % (k, v))
hdulist[0].header.update(match_header)
print('    writing "%s" (%d rows, %d columns) ...' % (outfile, len(tbhdu.data), len(columns)))
hdulist.writeto(outfile, **progress.kwargs_overwrite_true)

nwaylib.checkupdates.checkupdates()



"""Модуль с различными вариантами NWAY моделей."""

import pandas as pd
import numpy as np

import subprocess
import astropy
import os
from sklearn.metrics import roc_auc_score
from astropy.table import Table

from .base import Model
from .ml_models import Catboost
from .ml_models import ModelML
from ..metrics import Metric
from ...utils import local_package_path

from typing import Tuple, Union, Any


class NWAY(Model):
    """Модель NWAY для поиска отождествлений без использования приора."""

    def __init__(self, distance: float = 15, opt_err: float = 0.1, prior_completeness: float = 0.9) -> None:
        """Инициализация модели."""
        if not os.path.exists('nway_data'):
            os.mkdir('nway_data')
        self._distance = distance
        self._opt_err = opt_err
        self._prior_completeness = prior_completeness

    def _save_fits(self, df: pd.DataFrame, name: str, area: float) -> None:
        """Сохраняет DataFrame в FITS файл для применения его в NWAY."""
        data = Table.from_pandas(df)
        data = astropy.io.fits.table_to_hdu(data)
        data.header["SKYAREA"] = area
        data.header["EXTNAME"] = name
        data.writeto(f'nway_data/{name}.fits', overwrite=True)

    def _load_fits(self, name: str) -> pd.DataFrame:
        dat = Table.read(f'nway_data/{name}.fits', format='fits')
        df = dat.to_pandas()
        return df

    def _raname_opt(self, opt: pd.DataFrame, additional_columns: list[str] = []) -> pd.DataFrame:
        """Переименовывает столбцы входного каталога. Проверяет на наличие необходимых колонок: id, ra, dec."""
        for i in ['id', 'ra', 'dec']:
            if i not in opt.columns:
                raise ValueError(f'Column {i} is not in opt. Opt catalog should contain columns [id, ra, dec]')

        pan = opt[['id', 'ra', 'dec'] + additional_columns].rename(columns={'id': 'ID', 'ra': 'RA', 'dec': 'DEC'})
        return pan

    def _rename_xray(self, xray: pd.DataFrame, additional_columns: list[str] = []) -> pd.DataFrame:
        """Переименовывает столбцы входного каталога. Проверяет на наличие необходимых колонок: id, ra, dec, poserr."""
        for i in ['id', 'ra', 'dec', 'poserr']:
            if i not in xray.columns:
                raise ValueError(f'Column {i} is not in xray. Xray catalog should contain columns [id, ra, dec, poserr]')

        xmm = xray[['id', 'ra', 'dec', 'poserr'] + additional_columns].rename(columns={'id': 'ID', 'ra': 'RA', 'dec': 'DEC', 'poserr': 'pos_err'})
        return xmm

    def _split_data(self, df: pd.DataFrame, additional_opt: list[str] = [],
                    additional_xray: list[str] = []) -> Tuple[pd.DataFrame, pd.DataFrame]:
        """
        Разделяет каталог на оптические и рентгеновские каталоги.

        Args:
            df: Совмещенный каталог.
            additional_opt: Дополнительные столбцы оптического каталога, которые нужно оставить.
            additional_xray: Дополнительные столбцы рентгеновского каталога, которые нужно оставить.
        Returns:
            opt: Оптический каталог.
            xray: Рентгеновский каталог.
        """
        for i in ['id_xray', 'ra_xray', 'dec_xray', 'poserr', 'id_opt', 'ra_opt', 'dec_opt']:
            if i not in df.columns:
                raise ValueError(f'Column {i} is not in df. Catalog catalog should contain columns\
                                  [id_xray, ra_xray, dec_xray, poserr, id_opt, ra_opt, dec_opt]')

        xray = df[['id_xray', 'ra_xray', 'dec_xray', 'poserr'] + additional_xray].drop_duplicates()
        opt = df[['id_opt', 'ra_opt', 'dec_opt'] + additional_opt].drop_duplicates()
        xray.rename(columns={'id_xray': 'ID', 'ra_xray': 'RA', 'dec_xray': 'DEC', 'poserr': 'pos_err'}, inplace=True)
        opt.rename(columns={'id_opt': 'ID', 'ra_opt': 'RA', 'dec_opt': 'DEC'}, inplace=True)
        return opt, xray

    def _calc_area(self, opt: pd.DataFrame, xray: pd.DataFrame) -> float:
        """NotImplemented"""
        return 1283 / 30

    def get_predicted(self, df: Union[pd.DataFrame, None] = None, opt: Union[pd.DataFrame, None] = None,
                      xray: Union[pd.DataFrame, None] = None) -> pd.DataFrame:
        """Сделать предсказания модели с 3 вероятностями: p_i, P_i, P_0."""
        if df is None and opt is None and xray is None:
            raise ValueError('one of df or (opt and xray) should be not None')
        if df is not None and (opt is not None or xray is not None):
            raise ValueError('You should not pass df and (opt and xray) simultaneously')

        if df is not None:
            opt, xray = self._split_data(df)
        else:
            opt = self._raname_opt(opt)
            xray = self._rename_xray(xray)

        area = self._calc_area(opt, xray)
        self._save_fits(opt, 'opt', area)
        self._save_fits(xray, 'xray', area)

        nway_script_path = local_package_path('nway.py')
        subprocess.run(['python', nway_script_path, 'nway_data/xray.fits', ':pos_err', 'nway_data/opt.fits',
                        f'{self._opt_err}', '--out=nway_data/res_nway.fits', '--radius', f'{self._distance}',
                        '--prior-completeness', f'{self._prior_completeness}'], stdout=open(os.devnull, 'w'))

        df = self._load_fits('res_nway')
        return df

    def validate(self, X: pd.DataFrame, y: np.ndarray, metrics: Union[Metric, list[Metric]]) -> dict[Metric, float]:
        """Вычисляет значение переданных метрик для модели."""
        pred = self.get_predicted(X)
        X['mark'] = y
        tmp = pd.merge(X[['id_xray', 'id_opt', 'mark']], pred, left_on=['id_xray', 'id_opt'], right_on=['xray_ID', 'opt_ID'], how='inner')
        del X['mark']
        tmp['P_0'] = 1 - tmp['p_any']
        tmp['P_i'] = tmp['p_i']
        tmp = tmp[['id_xray', 'P_0', 'P_i', 'mark']]

        if not isinstance(metrics, list):
            res = {metrics: metrics.calculate(tmp)}
        else:
            res = dict()
            for metric in metrics:
                res[metric] = metric.calculate(tmp)

        return res


class NWAYauto(NWAY):
    """Модель NWAY для поиска отождествлений с использованием автоматически генерируемого приора."""

    def __init__(self, distance: float = 15, opt_err: float = 0.1, prior_completeness: float = 0.9, prior_radius: float = 3.5) -> None:
        """Инициализация модели."""
        super().__init__(distance, opt_err, prior_completeness)
        self.prior_radius = prior_radius

    def get_predicted(self, df: Union[pd.DataFrame, None] = None, opt: Union[pd.DataFrame, None] = None,
                      xray: Union[pd.DataFrame, None] = None) -> pd.DataFrame:
        """Сделать предсказания модели с 3 вероятностями: p_i, P_i, P_0."""
        if df is None and opt is None and xray is None:
            raise ValueError('one of df or (opt and xray) should be not None')
        if df is not None and (opt is not None or xray is not None):
            raise ValueError('You should not pass df and (opt and xray) simultaneously')

        if df is not None:
            self.nway_col = list(filter(lambda x: x.startswith('nway_'), df.columns))
            opt, xray = self._split_data(df, additional_opt=self.nway_col)
        else:
            opt = self._raname_opt(opt)  # Проверить надо ли убрать additional columns
            xray = self._rename_xray(xray)

        area = self._calc_area(opt, xray)
        self._save_fits(opt, 'opt', area)
        self._save_fits(xray, 'xray', area)

        nway_script_path = local_package_path('nway.py')
        script_list = ['python', nway_script_path, 'nway_data/xray.fits', ':pos_err', 'nway_data/opt.fits',
                       f'{self.opt_err}', '--out=nway_data/res_nway.fits', '--radius', f'{self.distance}',
                       '--prior-completeness', f'{self.prior_completeness}']

        for i in self.nway_col:
            print(i)
            script_list += ['--mag', f'opt:{i}', 'auto', '--mag-radius', f'{self.prior_radius}']

        subprocess.run(script_list)

        df = self._load_fits('res_nway')
        return df


class NWAYml(NWAY):

    def __init__(self, distance: float = 15, opt_err: float = 0.1, prior_completeness: float = 0.9,
                 ml_model: ModelML = Catboost(), bins: int = 20) -> None:
        """Инициализация модели."""
        super().__init__(distance, opt_err, prior_completeness)
        self.ml_model = ml_model
        self.bins = bins

    def save_prior(self, y: np.ndarray, pred: pd.DataFrame) -> None:
        """Сохранить приор, для использования его в модели."""
        arr = np.stack((y, pred)).T
        arr = np.array(sorted(arr, key=lambda x: x[1]))
        ans = []
        step = arr.shape[0] // self.bins
        for i in range(self.bins):
            mean_zeros = (arr[i * step:(i + 1) * step, 0] == 0).sum() / len(arr)
            mean_ones = (arr[i * step:(i + 1) * step, 0] == 1).sum() / len(arr)
            ans.append([arr[i * step, 1], arr[(i + 1) * step, 1], mean_ones, mean_zeros])
        ans = pd.DataFrame(ans, columns=['lo', 'hi', 'selected', 'others'])
        ans.to_csv('nway_data/prior.txt', sep='\t', index=False, header=False, float_format='%.4f')

    def fit(self, X: pd.DataFrame, y: np.ndarray) -> 'NWAYml':
        """Обучение модели."""
        self.prior_columns = list(filter(lambda x: x.startswith('nway_'), X.columns))
        self.ml_model.fit(X[self.prior_columns], y)
        print(self.prior_columns)
        pred = self.ml_model.predict_proba(X[self.prior_columns])
        print(roc_auc_score(y, pred))
        self.save_prior(y, pred)
        return self

    def predict_proba(self, X: pd.DataFrame) -> pd.DataFrame:
        return self.ml_model.predict_proba(X[self.prior_columns])

    def get_predicted(self, df: Union[pd.DataFrame, None] = None, opt: Union[pd.DataFrame, None] = None,
                      xray: Union[pd.DataFrame, None] = None) -> pd.DataFrame:
        """Сделать предсказания модели с 3 вероятностями: p_i, P_i, P_0."""
        if df is None and opt is None and xray is None:
            raise ValueError('one of df or (opt and xray) should be not None')
        if df is not None and (opt is not None or xray is not None):
            raise ValueError('You should not pass df and (opt and xray) simultaneously')

        # self.nway_col = list(filter(lambda x: x.startswith('nway_'), df.columns))  Проверить потом это

        if df is not None:
            self.nway_col = list(filter(lambda x: x.startswith('nway_'), df.columns))
            opt, xray = self._split_data(df, additional_opt=self.nway_col)
        else:
            opt = self._raname_opt(opt)
            xray = self._rename_xray(xray)

        area = self._calc_area(opt, xray)
        opt['MAG'] = self.predict_proba(opt)
        self._save_fits(opt, 'opt', area)
        self._save_fits(xray, 'xray', area)

        nway_script_path = local_package_path('nway.py')
        script_list = ['python', nway_script_path, 'nway_data/xray.fits', ':pos_err', 'nway_data/opt.fits',
                       f'{self.opt_err}', '--out=nway_data/res_nway.fits', '--radius', f'{self.distance}',
                       '--prior-completeness', f'{self.prior_completeness}', '--mag', 'opt:MAG', 'nway_data/prior.txt']

        subprocess.run(script_list)

        df = self._load_fits('res_nway')
        return df



"""Модуль, с моделями классификации, которые можно использовать для построения CustomModelContainer."""

from catboost import CatBoostClassifier
import pandas as pd
from sklearn.model_selection import train_test_split


class CustomCatboostModel(CatBoostClassifier):
    def __init__(self) -> None:
        super().__init__()

    def fit(self, X: pd.DataFrame, y: pd.Series) -> 'CustomCatboostModel':
        X_train, X_test, y_train, y_test = train_test_split(X, y, test_size=0.2, random_state=42)
        super().fit(X_train, y_train, eval_set=(X_test, y_test), use_best_model=True)
        return self



"""Модуль, реализующий оптимизацию гиперпараметров модели."""

from abc import abstractmethod
from typing import Tuple, Any
import pandas as pd
import numpy as np

from sklearn.model_selection import StratifiedKFold, cross_val_score
from functools import partial
from hyperopt import hp, fmin, tpe, Trials, STATUS_OK
import warnings


class ParamsOptimizator:
    """
    Класс для оптимизации параметров.

    Для построения собственного класса необходимо реализовать метод :func:`_params_space()`, в котором возвращаются \
    гиперпараметры модели в виде словаря c ключами из названия параметра и значениями из hyperopt.hp.
    """

    def _objective(self, params: dict[str, float], model: Any, X: pd.DataFrame, y: np.ndarray,
                   static_params: dict[str, Any], other_params: dict[str, Any], score_function: str) -> dict[str, Any]:
        """Функция для оптимизации."""
        model.set_params(params)
        model.set_params(static_params)
        model.set_params(other_params)
        skf = StratifiedKFold(n_splits=4, shuffle=True, random_state=42)
        score = cross_val_score(estimator=model, X=X, y=y, scoring=score_function, cv=skf, n_jobs=None)
        return {'loss': -score.mean(), 'params': params, 'status': STATUS_OK}

    def optimize(self, model: Any, X: pd.DataFrame, y: np.ndarray, n_iterations: int = 50, verbose: bool = True,
                 score_function: str = 'roc_auc', other_params: dict[str, Any] = {}) -> 'ParamsOptimizator':
        """Производит перебор гиперпараметров, из пространства поиска, предоставленного в методе param_space."""
        if self._params_space() is None:
            warnings.warn("You don't provide params space. Optimization skiped", Warning)
            return self

        params_space, static_params = self._params_space()  # type: ignore
        trials = Trials()
        best = fmin(fn=partial(self._objective, model=model, X=X, y=y, static_params=static_params,
                               other_params=other_params, score_function=score_function),
                    space=params_space, algo=tpe.suggest, max_evals=n_iterations, trials=trials,
                    show_progressbar=True, verbose=verbose)
        model.set_params(trials.best_trial['result']["params"])
        model.set_params(static_params)
        model.set_params(other_params)
        return self

    @abstractmethod
    def param_space(self) -> Tuple[dict, dict]:
        """
        Возвращает пространство для перебора гиперпараметров в виде словаря бибилиотеки hyperopt.

        Returns:
            dict: Словарь сетки для поиска гиперпараметров.
            dict: Словарь статических гиперпараметров, которые не будут меняться в ходе оптимизации.
        """
        pass


class CatBoostGPUOptimizator(ParamsOptimizator):
    """Класс для оптимизации CatBoost на GPU."""

    def param_space(self) -> Tuple[dict, dict]:
        """
        Возвращает пространство для перебора гиперпараметров с помощью бибилиотеки hyperopt.

        Returns:
            dict: Словарь сетки для поиска гиперпараметров.
            dict: Словарь статических гиперпараметров.
        """
        space = {'random_strength': hp.randint('random_strength', 1, 20),
                 'depth': hp.randint('depth', 1, 12),
                 'learning_rate': hp.uniform('learning_rate', np.exp(-7), 1),
                 'l2_leaf_reg': hp.uniform('l2_leaf_reg', 1, 10),
                 'bagging_temperature': hp.uniform('bagging_temperature', 0, 1),
                 'score_function': hp.choice(label='score_function', options=['L2', 'NewtonL2', 'NewtonCosine', 'Cosine']),
                 'leaf_estimation_backtracking': hp.choice(label='leaf_estimation_backtracking',
                                                           options=['No', 'AnyImprovement', 'Armijo']),
                 'leaf_estimation_iterations': hp.randint('leaf_estimation_iterations', 1, 20)}
        static = {
            'loss_function': 'CrossEntropy',
            'task_type': 'GPU',
            'devices': '0:1',
            'iterations': 100,
            'use_best_model': False,
            'logging_level': 'Silent'}

        return space, static


class CatBoostCPUOptimizator(ParamsOptimizator):
    """Класс для оптимизации CatBoost на CPU."""

    def param_space(self) -> Tuple[dict, dict]:
        """
        Возвращает пространство для перебора гиперпараметров с помощью бибилиотеки hyperopt.

        Returns:
            dict: Словарь сетки для поиска гиперпараметров.
            dict: Словарь статических гиперпараметров.
        """
        space = {'random_strength': hp.randint('random_strength', 1, 20),
                 'depth': hp.randint('depth', 1, 12),
                 'learning_rate': hp.uniform('learning_rate', np.exp(-7), 1),
                 'l2_leaf_reg': hp.uniform('l2_leaf_reg', 1, 10),
                 'bagging_temperature': hp.uniform('bagging_temperature', 0, 1),
                 'score_function': hp.choice(label='score_function', options=['L2', 'Cosine']),
                 'leaf_estimation_backtracking': hp.choice(label='leaf_estimation_backtracking',
                                                           options=['No', 'AnyImprovement']),
                 'leaf_estimation_iterations': hp.randint('leaf_estimation_iterations', 1, 20)}
        static = {
            'loss_function': 'CrossEntropy',
            'task_type': 'CPU',
            'iterations': 100,
            'use_best_model': False,
            'logging_level': 'Silent'}

        return space, static


class SklearnRFOptimizator(ParamsOptimizator):
    """Класс для оптимизации RandomForest из sklearn."""

    def param_space(self) -> Tuple[dict, dict]:
        """
        Возвращает пространство для перебора гиперпараметров с помощью бибилиотеки hyperopt.

        Returns:
            dict: Словарь сетки для поиска гиперпараметров.
            dict: Словарь статических гиперпараметров.
        """
        space = {'max_features': hp.choice(label='max_features', options=['log2', 'sqrt']),
                 'max_depth': hp.randint('max_depth', 10, 20),
                 'min_samples_split': hp.randint('min_samples_split', 1, 20),
                 'min_samples_leaf': hp.randint('min_samples_leaf', 1, 20),
                 'bootstrap': hp.choice(label='bootstrap', options=[True, False]),
                 'criterion': hp.choice(label='criterion', options=['gini', 'entropy', 'log_loss'])}
        static = {'n_jobs': -1,
                  'n_estimators': 1000}
        return space, static
